'''

Copyright 2004 Daniel J. Popowich
 
Licensed under the Apache License, Version 2.0 (the "License"); you
may not use this file except in compliance with the License.  You may
obtain a copy of the License at::

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied.  See the License for the specific language governing
permissions and limitations under the License.


See L{handler} for details about the servlet handler.

$Id: servlet.py,v 1.12 2005/02/12 04:10:00 popowich Exp $
'''

__all__ = ["Servlet", "HTMLPage", "handler"]

# Major, Minor, Patch
__version__ = 1, 1, 6
__version_str__ = '%d.%d.%d' % __version__

import sys, os, time, base64, re, urlparse
from mod_python import apache, util, Session, Cookie

# Extension expected for ModPython Servlet files.  This should contain
# a dot.
MPS = '.mps'

# A legal python identifier
NAME = r'[A-Za-z_]\w*'

# re for method calling identifier in form variable names:
# 
#    "_call_" NAME "(" ARGS ")"
#
# Group #1 is NAME, group #2 is ARGS.
METHODCALL = re.compile(r'_call_(%s)\((.*)\)$' % NAME)

# re for matching dictionaries in form variables:
#
#    NAME "[" KEY "]"
#
# Group #1 is NAME, group #2 is KEY
DICTVAR = re.compile(r'(%s)\[(.*)\]$' % NAME)

# re for matching form vars returned by INPUT tags of type image:
#
#    NAME "." [xy]
#
# Group #1 is NAME, group #2 is either 'x' or 'y'.
IMAGEVAR = re.compile(r'(%s)\.([xy])$' % NAME)

# Make a re out of NAME
NAME = re.compile(NAME + '$')

class Servlet:
    '''

    Abstract base class for all mod_python servlets.

    You must subclass this class implementing, minimally, L{respond},
    but you will probably want to subclass L{HTMLPage}, which already
    subclasses Servlet and has many added features for generating HTML
    output.

    @ivar content_type: Specifies the content type of the servlet,
    i.e., "text/html".  If it is not set (None) then it defaults to
    "text/plain".  Default: None

    @type content_type: str or None

    @ivar auth_realm: Specifies the realm for basic HTTP
    authentication.  Default: "Unspecified".

    @type auth_realm: str

    @ivar reusable: Flag (default: True) indicating whether or not an
    instance of this servlet can be used for multiple requests.  If
    False, instances will be recreated for every request of a servlet.

    @type reusable: bool

    @ivar instantiated: Timestamp of when the servlet was
    instantiated.  This is stored as seconds since the epoch; see the
    python documentation for time.time().
 
    @type instantiated: float

    @ivar req:  The apache request object.

    @ivar form: User data sent with request via form or url.  For each
    request an instance of FieldStorage is created.  Note:
    FieldStorage is created with blank values being ignored, ie, as if
    they did not exist.

    @type form: instance of mod_python.util.FieldStorage
    
    @ivar path_info: req.path_info canonicalized as a list: stripping
    beginning and trailing "/" and splitting it on internal "/".

    @type path_info: list of str

    @ivar session: A Session (see the mod_python documentation for
    mod_python.Session.Session).  A session object is created for each
    request if L{use_session} is True.  If L{use_session} is False,
    this variable will be set to None.

    @type session: instance of mod_python.Session.Session or None

    @ivar session_timeout: Length of time, in seconds, until session
    times out.  Default: 30 minutes.

    @type session_timeout: int

    @ivar ok_methods_to_call: list of methods that can be called
    directly via a POST.  If the name of a form element looks like::

       _call_NAME()

    and NAME is the name of an unbound method of the servlet and the
    method is listed in ok_methods_to_call, then that method will be
    called when the form is posted.  The form element can also look
    like this::

       _call_SOME_METHOD(1,2,3)

    and SOME_METHOD will be called with three arguments: 1, 2, 3.

    See the B{tutorial} that comes with the distribution for examples.

    @type ok_methods_to_call: list of unbound methods


    @ivar query_vars: List of arguments to be searched for in L{form}
    to be set as instance variables of the servlet.  This processing
    occurs during the call to L{prep}.

    query_vars can be either a list of strings (the variable names) or
    a list of tuples of the form::

      (NAME, DEFAULT [,CONVERSION])

    where NAME, a string, is the name of the variable and must be a
    legal python identifier; DEFAULT, must be a string, list or dict,
    is the default value for NAME if it does not appear in L{form},
    and, optionally, CONVERSION is a callable that can convert the
    string found in L{form} (or DEFAULT) to an appropriate value
    and/or type.

    When an element of query_var is just a string (not a tuple), this
    is equivalent to::

      (NAME, '')
      
    That is, the default value for NAME, if not found in L{form}, is
    an empty string.

    If DEFAULT is a string, the value will be retrieved from L{form}
    with a call to form.getfirst().  If DEFAULT is a list, the value
    will be retrieved by a call to form.getlist().  If DEFAULT is a
    dict, form will be searched for names with the pattern NAME[KEY]
    with calls to either getfirst() or getlist() depending on whether
    the default dict value is a list or string for that key.

    All string values retrieved from C{form} will be stripped of
    whitespace.

    Example of query_vars::

         ["name",
          ("login", "", bool),
          ("items", []),
          ("map", {"foo" : "bar",
                   "baz" : []})]

    query_vars should be contrasted to L{form_vars}.  Where
    C{form_vars} are only processed for POST requests, query_vars is
    processed for I{both} POST and GET requests.

    If installed, see the B{tutorial} for a live demonstration of
    query_vars.

    @ivar form_vars: This is the same as L{query_vars} except these
    variables are only processed for POST requests.  For I{all} GET
    requests, these variables will be set to their default values.

    @ivar use_session: If true, create (or reload) L{session} for each
    request.  Default: False.

    '''

    content_type = None
    auth_realm = 'Unspecified'
    reusable = True
    use_session = False
    session = None
    session_timeout = 60 * 30 # 30 minutes

    ok_methods_to_call = []

    query_vars = []
    form_vars = []

    def __init__(self):
        '''

        Base constructor for all servlets.

        If subclasses override this method they I{must} call this
        method for proper servlet initialization.  For example::

          class MyServlet(Servlet):

              def __init__(self):
                  Servlet.__init__(self)
                  ...

        '''

        self.instantiated = time.time()

        # create a dictionary for quick lookups of allowed methods to
        # call
        cache = {}
        for m in self.ok_methods_to_call:
            cache[m.__name__] = m
        self.__method_cache = cache

        # canonicalize query_vars and form_vars, checking the conversion
        # method, if specified, works on the default value
        for lst in self.query_vars, self.form_vars:
            assert type(lst) is list, 'query_vars and form_vars must be lists'
            for i, arg in enumerate(lst):
                lst[i] = self.__canonical_arg(arg)

    def __str__(self):
        '''
        
        Return a simple string representing an instance:

        "B{Servlet I{class name}}"

        '''
        return 'Servlet %s' % self.__class__.__name__

    def __canonical_arg(self, arg):
        '''

        Convert elements of L{query_vars} and L{form_vars} to a
        canonical form::

          (NAME, DEFAULT, CONVERSION)

        @param arg: an element of C{query_vars} or C{form_vars}

        '''
        
        if type(arg) is not tuple:
            arg = (arg, '', None)

        if len(arg) == 2:
            arg = arg + (None,)

        assert len(arg) == 3, ('Argument must be of the form: NAME or '
                               '(NAME, DEFAULT [,CONV])')
        name, default, conv = arg
        assert type(name) is str, 'Argument name must be a string'
        assert NAME.match(name), ('Argument name must be a legal '
                                  'python identifier')

        assert isinstance(default, (list, dict, str)), \
               ('Argument default value must be a string, '
                'list or dict (or subclass thereof)')
        
        if isinstance(default, list):
            for i in default:
                assert isinstance(i, str), ('Default values of lists must '
                                            'have elements of strings only')
        if isinstance(default, dict):
            for k,v in default.iteritems():
                assert type(k) is str, ('Default keys of dicts must '
                                        'be strings')
                assert isinstance(v, (str, list)), \
                       ('Default values of dicts must have values of lists '
                        'or strings')
                        
        assert conv is None or callable(conv), ('Argument conversion must '
                                                'be None or callable')
        # test conversion method works without exception on default
        if conv:
            try:
                conv(default)
            except Exception, msg:
                raise AssertionError, ('Argument conversion method failed '
                                       'on default value: %s' % msg)

        return name, default, conv


    def __load_vars(self):

        '''
        Set instance variables based on URI arguments and form variables.

        See the documentation for L{query_vars} and L{form_vars}.
        '''

        defaults_only = False

        for i, lst in enumerate((self.query_vars, self.form_vars)):

            # if we're processing form_vars and it's not a POST, use
            # defaults only
            if i == 1 and self.req.method != "POST":
                defaults_only = True

            for arg in lst:

                # break out the tuple
                name, default, conv = arg

                # The type of the default determines what we convert
                # from the form: list, dict or string

                if isinstance(default, list):
                    if defaults_only:
                        # use a copy of the default
                        val = default[:]
                    else:
                        # use the list of stripped strings
                        val = [x.strip() for x in self.form.getlist(name)]

                elif isinstance(default, dict):

                    # start with a copy of the dict, so any keys not
                    # in the form will have their default
                    val = default.copy()
                    if not defaults_only:
                        # iterate over every key in the form
                        for var in self.form.keys():
                            # see if it is a dict RE and the current name
                            m = DICTVAR.match(var) or IMAGEVAR.match(var)
                            if m and m.group(1) == name:
                                # get the key from the match
                                key = m.group(2)
                                if isinstance(val.get(key), list):
                                    # this item is a list
                                    val[key] = [x.strip() for x in
                                                self.form.getlist(var)]
                                else:
                                    # this item is a string
                                    val[key] = self.form.getfirst(var).strip()
                else:
                    # default is a string
                    if defaults_only:
                        val = default
                    else:
                        val = self.form.getfirst(name, default).strip()

                if conv:
                    try:
                        val = conv(val)
                    except:
                        val = conv(default)

                setattr(self, name, val)

    def __cleanup_session(self):
        '''

        Clean up session use.  If L{use_session} is True and
        L{session} is not None, save the session and delete it from
        the instance to allow for garbage collection.

        '''
        

        if self.use_session and self.session is not None:
            self.session.save()
            self.session = None

    def __canonical_path_info(self, pi):
        '''
        
        Private helper method to canonicalize the path info.

        Splits pi and returns path_info as list.
        
        @param pi: path_info

        @type pi: str

        '''

        if pi and pi != '/':
            last = pi[-1] == '/' and -1 or len(pi)
            pi = pi[1:last].split('/')
        else:
            pi = []

        return pi

    def _auth_with_mapping(self, mapping):
        '''

        Helper method to be used in L{auth}.  For example::

            def auth(self):
                self._auth_with_mapping(some_mapping)

        where I{some_mapping} is a mapping that supports the standard
        C{get} method and where the keys are usernames and their
        values passwords.
        
        '''

        user, pw = self._get_user_pw()
        if user and pw and mapping.get(user) == pw:
            return
        self._unauthorized()

    def _get_user_pw(self):
        '''
        
        Helper method to be used in L{auth} to retrieve username
        and password.

        @return: username, password
        @rtype: tuple of two strings
        '''

        if not self.req.headers_in.has_key('Authorization'):
            self._unauthorized()

        try:
            auth = self.req.headers_in['Authorization'][6:]
            user, pw = base64.decodestring(auth).split(':', 1)
        except:
            raise apache.SERVER_RETURN, apache.HTTP_BAD_REQUEST

        return user, pw


    def _unauthorized(self):

        '''
        
        Helper method to be used in L{auth}.

        Writes appropriate HTTP headers requesting authentication and
        raises an UNAUTHORIZED exception.

        '''

        self.req.err_headers_out["WWW-Authenticate"] = ('Basic realm="%s"'
                                                        % self.auth_realm)
        raise apache.SERVER_RETURN, apache.HTTP_UNAUTHORIZED

    def add_cookie(self, cookie, value="", **kw):
        '''

        Wrapper around mod_python.Cookie.add_cookie().  See mod_python
        documentation for details.

        '''

        Cookie.add_cookie(self.req, cookie, value, **kw)

    def get_cookies(self, klass=Cookie.Cookie, **kw):
        '''

        Wrapper around mod_python.Cookie.get_cookies().  See
        mod_python documentaion for details.

        '''

        return Cookie.get_cookies(self.req, klass, **kw)
    
    def external_redirect(self, uri, permanently=True):
        '''

        Send a redirect to the client via a B{Location} HTTP header.

        @param uri: The URI to relocate to.
        @type uri: str

        @param permanently: If True (the default) reply with 301
        (MOVED_PERMANENTLY, else 302 (MOVED_TEMPORARILY).
        @type permanently: bool


        @return: Does not return.

        '''

        self.req.headers_out["Location"] = uri
        moved = (permanently and apache.HTTP_MOVED_PERMANENTLY
                 or apache.HTTP_MOVED_TEMPORARILY)

        raise apache.SERVER_RETURN, moved

    def internal_redirect(self, uri):
        '''

        Redirect internally.  This does not send a redirect to the
        client (see L{external_redirect}), but redirects internally to
        the server.  This is a wrapper around req.internal_redirect();
        see mod_python documentation for details.

        @param uri: The URI to relocate to.
        @type uri: str

        @return: Does not return.

        '''
        

        # to prevent session deadlock
        self.__cleanup_session()

        self.req.internal_redirect(uri)

        raise apache.SERVER_RETURN, apache.OK

    def auth(self):
        '''
        
        Basic HTTP authentication method.

        This method is the first user method called by the L{handler}
        for each request, just before L{prep}.  It should return
        without exception if authorization is granted (the return
        value is ignored) or raise C{apache.SERVER_RETURN} with
        C{apache.HTTP_UNAUTHORIZED} as a value if authorization is
        denied.  Typical implementation should use the L{_get_user_pw}
        and L{_unauthorized} helper methods::

            def auth(self):
                user, pw = self._get_user_pw()
                # test user, pw for authorization; if OK, return
                self._unauthorized()

        See the source code for L{_auth_with_mapping} for a simple
        implementation.

        The base method is a no-op, i.e., no authentication is
        required.

        '''
        
        pass

    def prep(self):
        '''

        This is the second user method called by the L{handler}, after
        L{auth}, prior to L{respond}.

        It should be used as a means to prep the servlet for
        L{respond}, e.g., opening data files, acquiring db
        connections, preprocessing form data, etc.

        The return value is ignored by the handler.

        If this method is implemented in a subclass, the superclass
        method I{must} be called; e.g::

            class MyServlet(HTMLPage):
               ...

               def prep(self):
                   HTMLPage.prep(self)
                   ...
               ...

        '''

    
        self.__out = []

        self.form = util.FieldStorage(self.req)
        self.__load_vars()

        if self.use_session:
            self.session = Session.Session(self.req,
                                           timeout=self.session_timeout)

        # canonicalize the path_info, stripping beginning and trailing '/'
        # and turning it into a list (splitting it on '/')
        self.path_info = self.__canonical_path_info(self.req.path_info)

    def respond(self):
        '''

        This is the third user method called by the L{handler}, after
        L{prep}, before L{wrapup} and is where the response is
        generated.  Typically, this will be by calls to L{write},
        L{writeln} and/or L{raw_write}.  For example::

            def respond(self):
                self.writeln("Hello, world!")

        For most developers, this method will not need to be written
        because the version implemented in L{HTMLPage} is sufficient.

        The base class implementation (which is called by
        L{HTMLPage.respond<servlet.HTMLPage.respond>}) allows for
        arbitrary methods to be called during form POSTs.  See
        L{ok_methods_to_call}.

        This method must return True or False.  If True, request
        processing will continue (and ultimately, content sent to the
        client), if False, a 204 response will be sent to the client
        (NO_CONTENT) and processing will stop.

        @return: bool

        '''

        if self.__method_cache and self.req.method == 'POST':
            for key in self.form.keys():
                match = METHODCALL.match(key)
                if match:
                    method, args = match.group(1,2)
                    method = self.__method_cache.get(method)
                    if method:
                        args = args and args.split(',') or []
                        try:
                            return method(self, *args)
                        except TypeError:
                            # We'll assume that this TypeError
                            # indicates the method was called with the
                            # wrong number arguments, so we'll raise
                            # an appropriate exception.  Otherwise
                            # we'll let mod_python figure it out.
                            raise apache.SERVER_RETURN, apache.HTTP_BAD_REQUEST

        return False
    
    def wrapup(self):
        '''

        This is the fourth user method called by the L{handler},
        immediately after calling L{respond}, before L{_finally}.

        This method should be used to "tidy up" after a request, e.g.,
        flush output data, create a log entry, etc.

        The base method flushes output to the client.  If overridden,
        this method should be called by the subclass if it does not
        call L{flush}.

        The return value is ignored by the handler.

        '''
        
        self.flush()


    def _finally(self):
        '''
        
        This user method is the last called by the L{handler}.  The
        previously called user methods (L{auth}, L{prep}, L{respond},
        L{wrapup}) are called in a C{try} clause of a C{try...finally}
        statement.  This method is called in the C{finally} clause.

        This method should be used as a final "tidy up" after a
        request: in case of an exception this method is guaranteed to
        be called whereas L{wrapup} is not.

        The base method saves the session and removes it from the
        instance (thus implicitly unlocking the session) if
        L{use_session} is True.  If overridden, this method I{must} be
        called by the subclass.

        The output of this method is ignored by the handler.

        '''
        
        self.__cleanup_session()

        # in case there was an exception (and so output wasn't
        # flushed), set self.__out to None to garbage collect
        self.__out = None
    
    def log(self, msg):
        '''

        Utility method to write a message to the apache error log.

        @param msg: The message to be written to the log.
        @type msg: str or an object that can be converted into a str.

        @return: None

        '''
        print >> sys.stderr, ('[%s] mod_python Servlet [%s]: %s'
                              % (time.asctime(), self.__class__.__name__, msg))
        sys.stderr.flush()

    def write(self, *args):
        '''

        Utility method to write the arguments to the client.

        Each arg is coerced to a string and buffered for output.  The
        output is sent to the client when L{flush} is called
        explicitly or implicitly, when the handler is finished with
        the request.

        @param args: tuple of arbitrary objects
        @return: None

        '''
        for arg in args:
            self.__out.append(str(arg))
            
    def writeln(self, *args):
        '''

        Like L{write}, but in addition appends a newline to the output.
        
        @param args: tuple of arbitrary objects
        @return: None

        '''
        self.write(*args)
        self.write('\n')

    def raw_write(self, *args):
        '''

        Like L{write}, but all output is immediately flushed to the
        client and no string coercion is attempted on the arguments.

        @param args: tuple of arbitrary objects
        @return: None

        '''

        self.flush()
        
        for arg in args:
            self.req.write(arg)

    def flush(self):
        '''

        Utilily method which immediately flushes all buffered output
        to the client.

        If L{content_type} has not already been written it will be.

        @return: None

        '''
        if self.__out:
            if not self.req._content_type_set and self.content_type:
                self.req.content_type = self.content_type
                
            self.req.write(''.join(self.__out))
            self.__out = []


    def sourcefilename(klass):
        '''

        Get the filename of the source file for this servlet.

        @return: str

        '''

        return klass._sm_src
    sourcefilename = classmethod(sourcefilename)

    def count(klass):
        '''

        Get the count of requests for this servlet.

        @return: int

        '''

        return klass._sm_cnt
    count = classmethod(count)
    
    
class HTMLPage(Servlet):

    '''

    C{HTMLPage} is the servlet that most developers wanting to
    generate HTML will want to subclass.  The L{respond} method of
    C{HTMLPage} calls L{write_html} which writes to the client a
    well-defined HTML document.

    Many methods and instance variables offer features to the
    developer to allow the customization of their content.  Use of
    subclassing can produce site-wide continuity of content; e.g., you
    can create a servlet called C{SitePage} (that inherits from
    C{HTMLPage}) that will provide your site-wide look and feel, then
    individual servlets can inherit from C{SitePage}.

    To see how C{HTMLPage} writes out an HTML document, look at
    L{write_html} and the methods it calls.

    @ivar body_attrs: Attributes for BODY tag.  Default is empty dict.
    If non empty, the keys will become attribute names for the BODY
    tag and the values will be the attribute values.

    @ivar css_links: A list of hrefs (strings).  For each href, a
    <LINK type="text/css" rel="stylesheet" ...> will be placed in the
    HEAD.

    @ivar js_src: A list of hrefs (strings).  For each href, a <SCRIPT
    type="text/javascript" src="HREF"></SCRIPT> will be placed in the
    HEAD.

    @ivar shortcut_icon: If non empty, specifies the href to a
    shortcut icon and produces a LINK tag in the HEAD:

    <LINK rel="shortcut icon" href="I{shortcut_icon}">

    @ivar css: A string or list of strings of CSS.  If non empty, a
    <STYLE type="text/css">...</STYLE> section will be placed in the
    HEAD.  If css is a list of strings, the elements will be joined
    (seperated by newlines) and placed in a single <STYLE
    ...>...</STYLE> section.

    @ivar js: A string or list of strings of javascript.  If non
    empty, a <SCRIPT type="text/javascript">...</SCRIPT> section will
    be placed in the HEAD.  If js is a list of strings, the elements
    will be joined (seperated by newlines) and placed in a single
    <SCRIPT ...>...</SCRIPT> section.

    @ivar base: If set, this will specify the BASE tag in the HEAD.
    If C{base} is a string it will be the value of the C{href}
    attribute.  If C{base} is a tuple, it should have two string
    elements of the form (I{href}, I{target}) which will be the values
    of those BASE tag attributes.

    @ivar meta: A dict, which if non-empty will produce META tags in
    the HEAD.  The keys of the dict will be the values of the C{name}
    attribute and their values will become the C{content} attribute.
    For example, the following value of meta::

       meta = {"Author" : "Daniel Popowich",
               "Date"   : "April 28, 2004"}

    will produce the following output in the HEAD::

       <META name="Author" content="Daniel Popowich">
       <META name="Date" content="April 28, 2004">


    The values of the dict may be a list or tuple of strings, in which
    case multiple META tags will be produced for the same C{name}
    attribute.  For example::

       meta = {"Author" : ["Tom", "Jerry"]}

    will produce the following output in the HEAD::

       <META name="Author" content="Tom">
       <META name="Author" content="Jerry">

    @ivar http_equiv: http_equiv, like C{meta}, produces META tags in the
    HEAD, but instead of the keys being the value of the C{name}
    attribute, they are the value of the C{http-equiv} attribute.  See
    L{meta} for details.

    @ivar doctype: See L{write_doctype}
    @type doctype: str

    @ivar dtd: See L{write_doctype}
    @type dtd: one of "strict", "loose", "frameset"

    @ivar title: The title of the HTML page.  If not set, it defaults
    to the name of the servlet.  C{title} can be a callable that
    accepts no arguments, in which case the title will be the output
    of the callable.

    @type title: str or callable

    '''

    content_type = 'text/html'
    body_attrs = {}
    css_links = []
    js_src = []
    shortcut_icon = ''
    css = ''
    js = ''
    base = ''
    meta = {}
    http_equiv = {}

    def respond(self):
        '''

        This first calls L{Servlet.repsond<servlet.Servlet.respond>}
        (to see if a method should be called; see
        L{ok_methods_to_call<servlet.Servlet.ok_methods_to_call>}) and
        if it returns False, calls L{write_html}.

        '''
        
        return Servlet.respond(self) or self.write_html()

    def write_html(self):
        '''

        This method produces a well-formed HTML document and writes it
        to the client.

        It first calls L{write_doctype} which writes the DOCTYPE.  It
        then writes "<HTML>".  In turn it calls L{write_head} and
        L{write_body} which write the HEAD and BODY sections,
        respectively.  It finishes by writing "</HTML>".

        @return: True

        '''

        self.write_doctype()
        self.writeln('<html>')
        self.write_head()
        self.write_body()
        self.writeln('</html>')
        return True

    def write_doctype(self):
        '''

        Writes to the client the <!DOCTYPE...> tag.  By default, the
        following is used::

            <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
                      "http://www.w3.org/TR/html4/loose.dtd">        


        This can be controlled by setting the C{doctype} instance
        variable.  If set, it will be written, as-is, to the client.
        If it is not set, then the C{dtd} instance variable will be
        checked; it can be one of "strict", "loose" or "frameset".
        The default is "loose" and the resulting DOCTYPE is shown
        above.  If set to "strict" or "frameset" an appropriate
        DOCTYPE will be generated.

        This method will I{not} be typically overridden.

        '''
        

        try:
            doctype = self.doctype

        except AttributeError:

            # doctype not provided, so we'll generate it
            
            doctype = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01%s//EN"
            "http://www.w3.org/TR/html4/%s.dtd">'''

            dtds = {'strict' : '',
                    'loose'  : ' Transitional',
                    'frameset' : ' Frameset'}

            try:
                dtd = self.dtd

            except AttributeError:

                # DTD not provided, we'll assume the most liberal
                dtd = 'loose'
                
            if dtd not in dtds.keys():
                # if we were given an erroneous DTD, we'll quietly fix it
                dtd = 'loose'
                
            extra = dtds.get(dtd)

            # set doctype and cache it so we don't have to generate it again
            doctype = self.doctype = doctype % (extra, dtd)

        self.writeln(doctype)

    def write_head(self):
        '''

        This writes the HEAD section of the HTML document to the client.

        It first writes "<HEAD"> then calls L{write_head_parts} and,
        lastly, writes "</HEAD>".

        This method will I{not} be typically overridden.  To modify
        the content of the HEAD section, see L{write_head_parts}.

        '''

        self.writeln('<head>')
        self.write_head_parts()
        self.writeln('</head>')

    def write_head_parts(self):
        '''

        This method calls, in order:

          1. L{write_title}
          2. L{write_shortcut_icon}
          3. L{write_base}
          4. L{write_meta}
          5. L{write_css}
          6. L{write_js}

        See the above methods for details about what each writes to
        the client.  All of the above methods can have the content
        they write to the client controlled by setting (or unsetting)
        instance variables.

        While the base method will serve most developer needs, this
        method is a likely candidate to be overridden.  If you want to
        I{add} to the HEAD you will likely want to call the superclass
        method and then write additional content in your method,
        e.g.::

            class MyServlet(HTMLPage):
               ...

               def write_head_parts(self):
                   HTMLPage.write_head_parts(self)
                   # add my own content
                   self.writeln(...)
                   

        '''

        self.write_title()
        self.write_shortcut_icon()
        self.write_base()
        self.write_meta()
        self.write_css()
        self.write_js()

    def write_title(self):
        '''

        Writes the TITLE tag to the client.

        It checks for the C{title} instance variable.  If it does not
        exist it uses the name of the servlet for the title.  If
        C{title} is callable it will be called (with no arguments) and
        its output will be used as the title.

        @see: L{title}

        '''
        
        try:
            title = self.title
        except AttributeError:
            title = self.title = self.__class__.__name__

        if callable(title):
            title = title()
            
        self.writeln('<title>\n', title, '\n</title>')

    def __write_head_tag(self, tag, tagtype, content):
        'utility method to write HEAD tags'

        if type(content) in (list, tuple):
            content = '\n'.join(content)
                
        self.writeln('''
<%(tag)s type="%(tagtype)s">
<!--
%(content)s
-->
</%(tag)s>''' % locals())

    def write_shortcut_icon(self):
        '''

        Writes a LINK tag to the client specifying the shortcut icon.

        @see: L{shortcut_icon}

        '''

        if self.shortcut_icon:
            self.writeln('<link rel="shortcut icon" href="%s">'
                         % self.shortcut_icon)
            
    def write_css(self):
        '''

        Writes CSS to the client in the HEAD.

        @see: L{css_links}
        @see: L{css}

        '''

        for link in  self.css_links:
            self.writeln('<link type="text/css" rel="stylesheet" href="%s">'
                         % link)

        if self.css:
            self.__write_head_tag('style', 'text/css', self.css)

    def write_js(self):
        '''

        Writes javascript to the client in the HEAD.

        @see: L{js_src}
        @see: L{js}

        '''

        for src in self.js_src:
            self.writeln('<script src="%s" type="text/javascript"></script>'
                         % src)

        if self.js:
            self.__write_head_tag('script', 'text/javascript', self.js)

    def __write_meta(self, tag, mapping):
        'utility method to write META tags'

        for name, content in mapping.iteritems():

            if type(content) not in (list, tuple):
                content = (content,)

            for c in content:
                self.writeln('<meta %s="%s" content="%s">' % (tag, name, c))
                
            
    def write_meta(self):
        '''

        Writes META tags to the client in the HEAD.

        @see: L{meta}
        @see: L{http_equiv}

        '''
        
        if self.meta:
            self.__write_meta('name', self.meta)

        if self.http_equiv:
            self.__write_meta('http-equiv', self.http_equiv)

    def write_base(self):
        '''

        Write BASE tag to the client in the HEAD.

        @see: L{base}

        '''

        if not self.base:
            return

        if type(self.base) is tuple:
            href, target = self.base
        else:
            href, target = self.base, ''

        if href:
            href = 'href="%s"' % href
        if target:
            target = 'target="%s"' % target

        if href or target:
            self.writeln('<base %s %s>' % (href, target))
        
    
    def write_body(self):
        '''

        Writes the BODY section to the client.

        This method writes "<BODY...>" to the client taking into
        account the L{body_attrs} instance variable.  It then calls
        L{write_body_parts} and finally writes the "</BODY>" tag.

        This method will I{not} be typically overridden.  To modify
        the content of the BODY section, see L{write_body_parts}.

        '''

        body_attrs = []
        for attr, val in self.body_attrs.items():
            body_attrs.append('%s="%s"' % (attr, val))
        self.writeln('<body %s>' % ' '.join(body_attrs))
        self.write_body_parts()
        self.writeln('</body>')

    def write_body_parts(self):
        '''

        Write the content of the BODY section.

        The base implementation simply calls L{write_content}.

        For complex pages that have many sections, this method will
        typically be overridden, e.g., write to the client the layout
        of the site look and feel and where page-specific content
        should appear, call C{write_content}.

        See the B{tutorial} that comes with the distribution for
        examples of how this method can be used to create site-wide
        content.

        '''

        self.write_content()

    def write_content(self):
        '''

        This method must be overridden to produce content on the page.

        '''

        self.writeln('<strong>Override method <code>write_content'
                     '</code></strong>')

class _servlets_manager:

    def __init__(self):
        
        # keys are filenames, values a tuple: (class, mtime of file, servlets)
        self.class_cache = {}

    def __load_servlet(self, fname):

        '''Given a filename, return a Servlet class and cached
        instances, or raise an exception'''

        mtime = os.path.getmtime(fname)

        klass, cached_mtime, servlets = self.class_cache.get(fname,
                                                             (None,None, []))
        if not klass or mtime > cached_mtime:

            if klass:
                # we have a stale class; purge from cache
                del self.class_cache[fname]
                servlets = []
             
            # execute the code in fname
            code = {}
            execfile(fname, code)

            # look for the class defined as the same name as the file
            # (excluding, of course, the .mps extension)
            basename, ext = os.path.splitext(os.path.basename(fname))
            try:
                klass = code[basename]
            except:
                raise KeyError, 'class %s not found in %s' % (basename, fname)
            # make sure it is a subclass of Servlet
            try:
                if not issubclass(klass, Servlet):
                    raise ValueError
            except:
                raise ValueError, ('%s is not a subclass of Servlet (in %s)'
                                   % (basename, fname))

            # set the file from whence it came
            klass._sm_src = fname
            # initialize the count
            klass._sm_cnt = 0
            # cache self
            klass._sm = self
            # store the class, file's mtime and empty servlet list in the cache
            self.class_cache[fname] = klass, mtime, servlets

        # bump count
        klass._sm_cnt += 1
        return klass, servlets

    def verify_filename(self, req):
        '''
        Verify that the request is for a legal file and return its name.

        If the request specifies a directory, do a brute-force
        redirect to a file in the directory named I{index}.  This may
        result in a 404 error.

        File names beginning with a dot (.) or underscore (_) will
        result in a 403 error (FORBIDDEN).  This, in effect, reserves
        such files for internal use by an application (for example, to
        name database or configuration files).

        This function will return req.filename if it has an extension
        and the extension is not MPS (an empty string will be returned
        if req.filename has a MPS extension: this enforces clean URLS)
        or "%s%s" % (req.filename, MPS) if req.filename has no
        extension and the MPS file exists or req.filename in all other
        cases.
        
        @rtype: string

        '''

        # is it a directory?
        if os.path.isdir(req.filename):
            # brute force, try a redirect to an "index" page
            url = list(urlparse.urlsplit(req.unparsed_uri))
            url[2] = os.path.join(req.uri, 'index')
            req.internal_redirect(urlparse.urlunsplit(url))
            raise apache.SERVER_RETURN, apache.OK

        # files beginning with '.' or '_' will be protected for special use
        basename = os.path.basename(req.filename)
        if basename[0] in ('.', '_'):
            raise apache.SERVER_RETURN, apache.HTTP_FORBIDDEN

        # if the filename has an extension, return it.  EXCEPTION: mps
        # files, if specified with their extension will return an
        # empty string (resulting in a 404 error).  This enforces
        # clean URLs.
        root, ext = os.path.splitext(req.filename)
        if ext:
            return ext != MPS and req.filename or ''

        # the filename has no extension.  return FILENAME.mps if it
        # exists as a plain file, else FILENAME
        mps = '%s%s' % (req.filename, MPS)
        return os.path.isfile(mps) and mps or req.filename

    def get_servlet(self, filename):
        'Given a mps filename, return an instance of Servlet'

        klass, servlets = self.__load_servlet(filename)

        try:
            servlet = servlets.pop()
        except IndexError:
            servlet = klass()

        return servlet

    def put_servlet(self, servlet):
        'Put a servlet back into the cache for reuse.'

        # get the class and servlet list from the cache
        klass, mtime, servlets = self.class_cache.get(servlet._sm_src,
                                                      (None, None, None))
        # it's possible (but extremely unlikely) that the class has
        # since been purged, so check before putting it back
        if klass == servlet.__class__:
            servlets.append(servlet)

_sm = _servlets_manager()

def handler(req):

    """This handler uses instances of subclasses of L{Servlet} to
    handle HTTP GET and POST requests.  These instances are referred
    to as I{servlets} in this documentation.

    Overview of how it works:

      1. A developer creates a class that subclasses L{Servlet} or,
         more probably, L{HTMLPage}, which already subclasses
         L{Servlet} and has many features for generating HTML.

      2. Let's say we have a class named C{Foo} that subclasses
         L{HTMLPage}.  This class must be saved in a file called
         Foo.mps.  (mps stands for Mod Python Servlet).

         Generally, for any servlet named SERVLET, it must be a
         subclass of L{Servlet} and saved in a file named SERVLET.mps.
         

      3. Apache needs to be configured to call this handler for
         Foo.mps.  Assuming /var/www/html is your DocumentRoot and
         also assuming Foo.mps is saved in directory
         /var/www/html/mps_test then Apache needs the following
         configuration::

            <Directory /var/www/html/mps_test>
              SetHandler mod_python
              PythonHandler mod_python.servlet
              PythonDebug on
            </Directory>

      4. When the server receives a request for I{/mps_test/Foo} it will
         look for a file named Foo.mps in /var/www/html/mps_test, compile
         the file, look for a class named Foo, insure it is a subclass
         of L{Servlet}, create an instance of this class, let's call
         it C{servlet}, and then call, in order:

           1.  C{servlet.L{auth<servlet.Servlet.auth>}()}
           2.  C{servlet.L{prep<servlet.Servlet.prep>}()}
           3.  C{servlet.L{respond<servlet.Servlet.respond>}()}
           4.  C{servlet.L{wrapup<servlet.Servlet.wrapup>}()}


         See the documentation for the above methods of L{Servlet} for
         details about each stage of processing the request.

      5. Instances of a class are cached, so if a request for
         I{/mps_test/Foo} comes in another request, the instance will
         be found in the cache and reused.


    See the documentation for L{Servlet} and L{HTMLPage} for further
    details about servlet processing and helpful features.

    If installed, see the B{tutorial} for a live demonstration of
    servlets.  See the README file that came with this distribution
    for tutorial installation instructions.

    """

    # This handler only processes GET and POST requests
    if req.method not in ["GET", "POST"]:
        req.allow_methods(["GET", "POST"])
        return apache.HTTP_METHOD_NOT_ALLOWED

    # get the filname which is either an apache-configured
    # UberServlet (one servlet to handle all requests) or possibly a
    # modpython servlet.
    filename = req.get_options().get('UberServlet') or _sm.verify_filename(req)

    # if the file doesn't exist return 404
    if not os.path.isfile(filename):
        return apache.HTTP_NOT_FOUND

    # if it's not a mps file, decline the request
    base, ext = os.path.splitext(filename)
    if ext != MPS:
        return apache.DECLINED
    
    # get the servlet
    servlet = _sm.get_servlet(filename)

    # stuff request in servlet
    servlet.req = req

    # call user methods
    try:
        servlet.auth()
        servlet.prep()
        if not servlet.respond():
            return apache.HTTP_NO_CONTENT
        servlet.wrapup()

    finally:

        # handler wrapup of servlet
        servlet._finally()

        # if the servlet is reusable, put it back into the cache
        if servlet.reusable:
            _sm.put_servlet(servlet)

    return apache.OK
